-- Copyright University of Bergen, 2009
-- Written by Dag Hovland, july 2009


module Inclusion 
-- (Regexp, Regexp', first, headerForm, runInclusion, parseRE, Inclusion, inclusionTA, Inclusion.reverse, aux_prop_inc_refl) 
where

import Control.Monad
import List
import Data.Maybe
import Test.QuickCheck
import Data.Char


-- BNF re ::= l | re + re | ( re ) | re re | re*

data Regexp = Letter Char | Choice Regexp Regexp | Cat Regexp Regexp | Inter [Regexp] | Star Regexp | Epsilon 
            deriving (Eq, Ord)

--instance Eq Regexp where
--    (==) r1 r2 = sameRE (headerForm r1) (headerForm r2)


instance Show Regexp where
    show re = remChars (foldRE 
                        "$e"
                        (\l -> [l])
                        (\r ->  ('(':r) ++ ")*")
                        (\r1 -> \r2 -> (r1 ++ " " ++ r2))
                        (\rl -> "(" ++ (foldr (\r -> \s -> r ++ "&" ++ s) "" rl) ++ ")")
                        (\r1 -> \r2 -> ("(" ++ r1 ++ "+" ++ r2 ++ ")"))
                        re
                       )
              

-- for showing a tex-version of the regexp
texShow :: Regexp -> String
texShow Epsilon = "\\epsilon"
texShow (Letter l) = [l]
texShow (Star (Cat r1 r2)) = "(" ++ texShow (Cat r1 r2) ++ ")^*"
texShow (Star r) = texShow r ++ "^*"
texShow (Cat r Epsilon) = texShow r
texShow (Cat r1 r2) = texShow r1 ++ " " ++ texShow r2
texShow (Inter rl) = "(" ++ (foldr (\r s -> texShow r ++ "&" ++ s) "" rl) ++ ")"
texShow (Choice r1 r2) = "(" ++ texShow r1 ++ "+" ++ texShow r2 ++ ")"


remChars :: String -> String
remChars [] = []
remChars (c1:rs) = if c1 == '"' || c1 == '\\' then remChars rs else c1:(remChars rs)

isLetter :: Regexp -> Bool
isLetter (Letter _) = True
isLetter _ = False

--instance Eq Regexp where 
--    (==) r1 r2 = sameRE (headerForm r1) (headerForm r2)

sameRE :: Regexp -> Regexp -> Bool
sameRE Epsilon Epsilon = True
sameRE (Letter l) (Letter m) = l == m 
sameRE (Choice r1 r2) (Choice r3 r4) = sameRE r1 r3 && sameRE r2 r4
sameRE (Cat r1 Epsilon) r2 = sameRE r1 r2
sameRE r1 (Cat r2 Epsilon) = sameRE r1 r2
sameRE (Cat Epsilon r1) r2 = sameRE r1 r2 
sameRE r1 (Cat Epsilon r2) = sameRE r1 r2 
sameRE (Cat r1 r2) (Cat r3 r4) = sameRE r1 r3 && sameRE r2 r4 
sameRE (Star r1) (Star r2) = sameRE r1 r2
sameRE (Inter rl1) (Inter rl2) = length rl1 == length rl2 && foldr (\(r1,r2) s -> s && sameRE r1 r2) True (zip  rl1 rl2)
sameRE _ _ = False

foldRE :: a -> (Char -> a) -> (a -> a) -> (a -> a -> a) -> ([a] -> a) -> (a -> a -> a) -> Regexp -> a
foldRE eps _ _ _ _ _ Epsilon = eps
foldRE _ chr _ _ _ _ (Letter p) = chr p
foldRE eps chr str cat int cho (Star r) = str (foldRE eps chr str cat int cho r)
foldRE eps chr str cat int cho (Cat r1 r2) = cat (foldRE eps chr str cat int cho r1) (foldRE eps chr str cat int cho r2)
foldRE eps chr str cat int cho (Choice r1 r2) = cho (foldRE eps chr str cat int cho r1) (foldRE eps chr str cat int cho r2)
foldRE eps chr str cat int cho (Inter rl) = int (foldr (\el ls -> ((foldRE eps chr str cat int cho el) : ls)) [] rl)

mapRE :: (Regexp -> Regexp) -> Regexp -> Regexp
mapRE _ Epsilon = Epsilon
mapRE _ (Letter l) = Letter l
mapRE f (Star r) = Star (f (mapRE f r))
mapRE f (Cat r1 r2) = (Cat (f (mapRE f r1)) (f (mapRE f r2)))
mapRE f (Choice r1 r2) = (Choice (f (mapRE f r1)) (f (mapRE f r2)))
mapRE f (Inter rl) = Inter (map (\r -> f (mapRE f r)) rl)


instance Arbitrary Char where
    arbitrary     = choose ('\32', '\128')
    coarbitrary c = variant (Data.Char.ord c `rem` 4)


--    arbitrary = do 
--      l <- choose ('a','c') :: Gen Char
--      n <- choose (1,5) :: Gen Int
--      case n of
--           1 -> return Epsilon
--
--           2 -> do r1 <- arbitrary
--                   r2 <- arbitrary
--                   return (Cat r1 r2) 
--
--           3 -> do r1 <- arbitrary
--                   r2 <- arbitrary
--                   return (Choice r1 r2) 
--
--           4 -> do r1 <- arbitrary
--                   return (if nullable r1 then (Star (Cat (Letter l) r1)) else (Star r1))
--
--           5 -> do return (Letter l)

instance Arbitrary Regexp where
    arbitrary = sized tree'
        where tree' :: Int -> Gen Regexp
              tree' 0 = do l <- choose ('a','c') :: Gen Char
                           oneof [ return Epsilon, return (Letter l) ]
              tree' n | n > 0 =
                          do l <- choose ('a','c') :: Gen Char
                             oneof [ return Epsilon
                                   , return (Letter l)
                                   , liftM2 Cat subexpr subexpr
                                   , liftM2 Choice subexpr subexpr
                                   , liftM Star subexpr1 ]
                                 where subexpr = tree' (n `div` 2)
                                       subexpr1 = tree' (n - 1)



        
starNormalForm :: Regexp -> Bool
starNormalForm re = isEpsilonForm re &&
                    snd (foldRE
                         (Epsilon, True)
                         (\c -> (Letter c, True))
                         (\(r, b) -> (Star r, b && not (nullable r)))
                         (\(r1,b1) (r2,b2) -> (Cat r1 r2, b1 && b2))
                         (\rl -> foldr (\(r1,b1) -> \(Inter rl, b2) -> (Inter (r1:rl), b1 && b2)) (Inter [], True) rl)
                         (\(r1,b1) (r2,b2) -> (Choice r1 r2, b1 && b2))
                         (epsilonForm re) )                        
-- Parser
--Functional parsing library from chapter 8 of Programming in Haskell,
--Graham Hutton, Cambridge University Press, 2007.



--The monad of parsers
--------------------

data Parser a              =  P (String -> [(a,String)])

parse                         :: Parser a -> String -> [(a,String)]
parse (P p) inp               =  p inp


instance Monad Parser where
    return v                   =  P (\inp -> [(v,inp)])
    p >>= f                    =  P (\inp -> case parse p inp of
                                                []        -> []
                                                [(v,out)] -> parse (f v) out)
  

instance MonadPlus Parser where
    mzero                      =  P (\_ -> [])
    p `mplus` q                =  P (\inp -> case parse p inp of
                                               []        -> parse q inp
                                               [(v,out)] -> [(v,out)])

choice :: forall a. Parser a -> Parser a -> Parser a
p `choice`q = P (\inp -> parse p inp ++ parse q inp)

--Basic parsers
-------------

failure                       :: Parser a
failure                       =  mzero

item                          :: Parser Char
item                          =  P (\inp -> case inp of
                                              []     -> []
                                              (x:xs) -> [(x,xs)])
 



--Choice
------


(+++)                         :: Parser a -> Parser a -> Parser a
p +++ q                       =  p `mplus` q


--Derived primitives
------------------


sat                           :: (Char -> Bool) -> Parser Char
sat p                         =  do x <- item
                                    if p x then return x else failure

lower                         :: Parser Char
lower                         =  sat isLower
 
upper                         :: Parser Char
upper                         =  sat isUpper



letter                        :: Parser Char
letter                        =  sat isAlpha


alphanum                      :: Parser Char
alphanum                      =  sat isAlphaNum

 
parseChar                          :: Char -> Parser Char
parseChar x                        =  sat (== x)

 
string                        :: String -> Parser String
string []                     =  return []
string (x:xs)                 =  do parseChar x
                                    string xs
                                    return (x:xs)

many                          :: Parser a -> Parser [a]
many p                        =  many1 p +++ return []


many1                         :: Parser a -> Parser [a]
many1 p                       =  do v  <- p
                                    vs <- many p
                                    return (v:vs)

isSpace :: Char -> Bool
isSpace c = elem c " \t\n\\"

space                         :: Parser ()
space                         =  do many (sat Inclusion.isSpace)
                                    return ()

--Ignoring spacing
----------------

token                         :: Parser a -> Parser a
token p                       =  do space
                                    v <- p
                                    space
                                    return v



symbol                        :: String -> Parser String
symbol xs                     =  token (string xs)

expr                          :: Parser Regexp
expr                          =  do space
                                    t <- term
                                    do space
                                       symbol "+"      
                                       space
                                       e <- expr
                                       return (Choice t e)
                                     +++ return t


term                          :: Parser Regexp
term                          =  do space
                                    f <- starexp
                                    do space
                                       t <- term
                                       return (Cat f t)
                                     +++ return f

starexp                        :: Parser Regexp
starexp                        =  do e <- factor
                                     do symbol "*"
                                        return (Star e)
                                      +++ return e

factor                        :: Parser Regexp
factor                        =  do space
                                    symbol "("
                                    space
                                    e <- expr
                                    space
                                    symbol ")"
                                    return e
                                +++ epsilon
                                +++ letterexp


letterexp                        :: Parser Regexp
letterexp                        = do e <- alphanum
                                      return (Letter e)


epsilon                         :: Parser Regexp
epsilon                         = do parseChar '$'
                                     parseChar 'e'
                                     return Epsilon



parseRE                         :: String -> Regexp
parseRE xs                      = case (parse expr xs) of
                                    [(n,[])] -> n
                                    [(_,out)] -> error ("unused input " ++ out)
                                    [] -> error "invalid input"
        

-- set operation
setunion :: (Eq a) => [a] -> [a] -> [a]
setunion s1 s2 = s1 ++ [ x | x <- s2, not(elem x s1)]

subset :: (Eq a) => [a] -> [a] -> Bool
subset [] _ = True 
subset (el:tl) s2 = if (elem el s2) then (subset tl s2) else False


hasIntersection :: (Eq t) => [t] -> [t] -> Bool
hasIntersection [] _ = False
hasIntersection (hd:tl) s2 = if (elem hd s2) 
                             then True 
                             else hasIntersection tl s2 




-- Nullable
nullable :: Regexp -> Bool
nullable Epsilon = True
nullable (Star _) = True
nullable (Letter _) = False
nullable (Choice r1 r2) = (nullable r1) || (nullable r2) 
nullable (Inter []) = True
nullable (Inter rl) = List.or (map nullable rl)
nullable (Cat r1 r2) = (nullable r1) && (nullable r2) 


emptyLang :: Regexp -> Bool
emptyLang re = foldRE True (\_ -> False) id 
               (\r1 r2 -> r1 && r2)
               (\rl -> foldr (\r -> \b -> b && r) True rl)
               (\r1 r2 -> r1 && r2) re
-- First-set

firstset :: Regexp -> [Char]
firstset Epsilon = []
firstset (Letter l) = [l]
firstset (Star r) = firstset r
firstset (Cat r1 r2) = let f1 = firstset r1 in
                       if (nullable r1) then setunion f1 (firstset r2) 
                       else (firstset r1)
firstset (Inter rl) = foldr (\r s -> setunion (firstset r) s) [] rl
firstset (Choice r1 r2) = setunion (firstset r1) (firstset r2)

headerFormExtra :: Regexp -> Regexp
headerFormExtra re = headerForm (headerForm_aux re)

-- header form
headerForm :: Regexp -> Regexp
headerForm Epsilon = Epsilon
headerForm (Choice r1 r2) = (Cat (Choice r1 r2) Epsilon)
headerForm (Star r1) = (Cat (Star r1) Epsilon)
headerForm (Inter rl) = (Cat (Inter rl) Epsilon)
headerForm (Letter l) = (Cat (Letter l) Epsilon)
headerForm (Cat Epsilon r1) = headerForm r1
headerForm (Cat (Cat r1 r2) r3) = headerForm (Cat r1 (Cat r2 r3))
headerForm (Cat (Choice r1 r2) r3) = (Cat (Choice r1 r2) r3)
headerForm (Cat (Star r1) r2) = (Cat (Star r1) r2)
headerForm (Cat (Inter rl) r2) = (Cat (Inter rl) r2)
headerForm (Cat (Letter l) r1) = (Cat (Letter l) r1)

headerForm_aux :: Regexp -> Regexp
headerForm_aux Epsilon = Epsilon
headerForm_aux (Letter l) = Letter l
headerForm_aux (Star r) = Star (headerForm_aux r)
headerForm_aux (Cat r Epsilon) = headerForm_aux r
headerForm_aux (Cat Epsilon r) = headerForm_aux r
headerForm_aux (Cat r1 r2) = Cat (headerForm_aux r1) (headerForm_aux r2)
headerForm_aux (Inter rl) = Inter (foldr (\r s -> (headerForm_aux r) : s) [] rl)
headerForm_aux (Choice r1 r2) = Choice (headerForm_aux r1) (headerForm_aux r2)

-- Makes all epsilon subexpressions to epsilon
epsilonForm :: Regexp -> Regexp
epsilonForm (Star r1) | emptyLang r1 = Epsilon
                      | otherwise = Star (epsilonForm r1)
epsilonForm (Cat r1 r2) | emptyLang r1 && emptyLang r2 = Epsilon
                        | otherwise = Cat (epsilonForm r1) (epsilonForm r2)
epsilonForm (Choice r1 r2) | emptyLang r1 && emptyLang r2 = Epsilon
                           | otherwise = Choice (epsilonForm r1) (epsilonForm r2)
epsilonForm (Inter []) = Epsilon 
epsilonForm (Inter rl) = let ls = (foldr (\r -> \ls -> if r == Epsilon then ls else epsilonForm r : ls) [] rl) in
                         if ls == [] then Epsilon else Inter ls
epsilonForm r = r

      
isEpsilonForm :: Regexp -> Bool
isEpsilonForm re = snd (foldRE
                        (Epsilon, True)
                        (\c -> (Letter c, True))
                        (\(r, b) -> (Star r, b && not (emptyLang r)))
                        (\(r1, b1) (r2, b2) -> ((Cat r1 r2), 
                                                (r1 /= Epsilon || r2 /= Epsilon)
                                                && b1 && b2))
                        (foldr (\(r,b1) -> \(Inter s, b2) -> (Inter (r:s), b1 && b2 && r /= Epsilon)) (Inter [], True))
                         (\(r1, b1) (r2, b2) -> ((Choice r1 r2), 
                                                (r1 /= Epsilon || r2 /= Epsilon)
                                                 && b1 && b2))
                         re )
                   
                   


                            


-- Derivation Tree
data Inclusion = Axm Regexp 
               | LetterRule Char Regexp Regexp Inclusion 
               | LetterStar Char Regexp Regexp Regexp Inclusion 
               | LetterChoice2 Char Regexp Regexp Regexp Regexp Inclusion 
               | LetterChoice3 Char Regexp Regexp Regexp Regexp Inclusion 
               | LeftChoice Regexp Regexp Regexp Regexp Inclusion Inclusion 
               | LeftStar Regexp Regexp Regexp Regexp Inclusion Inclusion
               | StarStarE Regexp Regexp Inclusion
               | StarChoice1 Regexp Regexp Regexp Regexp Regexp Inclusion
               | StarChoice2 Regexp Regexp Regexp Regexp Regexp Inclusion Inclusion
               | ElimCat Regexp Regexp Regexp Inclusion
               | Loop Regexp Regexp
               | Not1Unambiguous String
               | NotInclusion Regexp Regexp String
                 deriving (Show, Eq)
               
  

data Included = Included | OneAmbiguous | NotIncluded
              deriving (Show, Eq, Read)

aux_prop_inc_refl :: Inclusion -> Bool
aux_prop_inc_refl (NotInclusion _ _ _) = False
aux_prop_inc_refl _ = True

-- Counts number of lines in a string
countLines :: Num a => String -> a
countLines "" = 0
countLines ('\n':xs) = 1 + countLines xs
countLines (_:xs) = countLines xs

-- Returns lines separated by \n as array of lines
getLines :: String -> [String]
getLines s = getLines' "" s

getLines' :: String -> String -> [String]
getLines' s ('\n':ss) = (List.reverse s) : (getLines' "" ss)
getLines' s "" = [s]
getLines' s (c:ss) = getLines' (c:s) ss 


--instance Show Inclusion where 
--    show (Axm r) = "\n-------------\nAxm: " ++ (show Epsilon) ++ " <= " ++ (show r) ++ "\n"
--    show (LetterRule c r1 r2 i) = (show i) ++ "Letter: " ++ (show c) ++ (show r1) ++ " <= " ++ (show c) ++ (show r2) ++ "\n"
--    show (LetterStar c r1 r2 r3 i) = (show i) ++ "LetterStar: " ++ (show c) ++ (show r1) ++ " <= " ++ (show (Cat (Star r2) r3)) ++ "\n"
--    show (LetterChoice2 c _ _ _ _ i) = (show i) ++ "LetterChoice2: " ++ (show c) ++ "\n"
--    show (LetterChoice3 c _ _ _ _ i) = (show i) ++ "LetterChoice3: " ++ (show c) ++ "\n"
--    show (LeftChoice _ _ _ _ i1 i2) = (show i1) ++ (show i2) ++ "\nLetterChoice3\n"
--    show (StarStar1 r1 r2 r3 r4 i1 i2) = (show i1) ++ (show i2) ++ "StarStar1: " ++ (show (Cat (Star r1) r2)) ++ "<=" ++ (show (Cat (Star r3) r4)) ++ ": r1=" ++ (show r1) ++ ", r2=" ++ (show r2) ++ ", r3=" ++ (show r3) ++ " and r4=" ++ (show r4) ++ "\n"
--    show (StarStar2 _ _ _ _ i1) = (show i1)  ++ "StarStar2\n"
--    show (StarChoice1 _ _ _ _ _ i1) = (show i1)  ++ "StarChoice1\n"
--    show (StarChoice2 _ _ _ _ _ i1 i2) = (show i1)  ++ "" ++ (show i2) ++ "\nStarChoice1\n"
--    show (ElimCat _ r2 _ i1) = (show i1)  ++ "\nElimCat, elim: \"" ++ (show r2) ++"\"\n"
--    show (Loop r1 r2) = "\nLooping on \"" ++ (show r1) ++ "\" <= \"" ++ (show r2) ++ "\n"
--    show (Not1Unambiguous s) = "Not 1-unambiguous: " ++ s ++ "\n"
--    show (NotInclusion s) = "\nNot in inclusion relation: " ++ s ++ "\n"

prettyIncl :: Inclusion -> String
prettyIncl (Axm r) = "\n-------------\nAxm: " ++ (show Epsilon) ++ " <= " ++ (show r) ++ "\n"
prettyIncl (LetterRule c r1 r2 i) = (prettyIncl i) ++ "Letter: " ++ (show c) ++ (show r1) ++ " <= " ++ (show c) ++ (show r2) ++ "\n"
prettyIncl (LetterStar c r1 r2 r3 i) = (prettyIncl i) ++ "LetterStar: " ++ (show c) ++ (show r1) ++ " <= " ++ (show (Cat (Star r2) r3)) ++ "\n"
prettyIncl (LetterChoice2 c _ _ _ _ i) = (prettyIncl i) ++ "LetterChoice: i=2, " ++ (show c) ++ "\n"
prettyIncl (LetterChoice3 c _ _ _ _ i) = (prettyIncl i) ++ "LetterChoice: i=3, " ++ (show c) ++ "\n"
prettyIncl (LeftChoice r1 r2 r3 r4 i1 i2) = (prettyIncl i1) ++ (prettyIncl i2) ++ "\nLeftChoice: " ++ (show (Cat (Choice r1 r2) r3)) ++ " <= " ++ (show r4) ++ "\n"
prettyIncl (LeftStar r1 r2 r3 r4 i1 i2) = (prettyIncl i1) ++ (prettyIncl i2) ++ "LeftStar " ++ (show (Cat (Star r1) r2)) ++ " <= " ++ (show (Cat r3 r4)) ++ ": r1=" ++ (show r1) ++ ", r2=" ++ (show r2) ++ ", r3=" ++ (show r3) ++ " and r4=" ++ (show r4) ++ "\n\n"
prettyIncl (StarStarE r1 r2 i1) = (prettyIncl i1) ++ "StarStarE: " ++ (show (Star r1)) ++ " <= " ++ (show (Star r2)) ++ ": r1=" ++ (show r1) ++ ", r2=" ++ (show r2) ++ "\n\n"
prettyIncl (StarChoice1 r1 r2 r3 r4 r5 i1) = (prettyIncl i1)  ++ "StarChoice1: " ++ (show (Cat (Star r1) r2)) ++ " <= " ++ (show (Cat (Choice r3 r4) r5)) ++ "\n"
prettyIncl (StarChoice2 r1 r2 r3 r4 r5 i1 i2) = (prettyIncl i1)  ++ "" ++ (prettyIncl i2) ++ "\nStarChoice2: " ++ (show (Cat (Star r1) r2)) ++ " <= " ++ (show (Cat (Choice r3 r4) r5)) ++ "\n"
prettyIncl (ElimCat _ r2 _ i1) = (prettyIncl i1)  ++ "\nElimCat, elim: \"" ++ (show r2) ++"\"\n"
prettyIncl (Loop r1 r2) = "\nLooping on \"" ++ (show r1) ++ "\" <= \"" ++ (show r2) ++ "\n"
prettyIncl (Not1Unambiguous s) = "Not 1-unambiguous: " ++ s ++ "\n"
prettyIncl (NotInclusion r1 r2 s) = "\nNot in inclusion relation: " ++ s ++ "\n"

-- auxiliary for texIncl
showRuleName :: String -> String
showRuleName "" = ""
showRuleName r = "\\mathsf{(" ++ r ++ ")} \\; \\;"

-- Prints out a tex version of the decision tree
-- The bool decides whether to print rule names,
-- The string argument is the optional rule name
-- The Inclusion is the decision tree
texIncl :: Bool -> String -> Inclusion -> String
texIncl v rName (Axm r) = ("\\dfrac{\\funcname{(Axm)}}{" ++ showRuleName rName  ++ (texShow Epsilon) ++ "\\sqsubseteq " ++ (texShow r) ++ "}")
texIncl v rName (LetterRule c r1 r2 i) = "\\dfrac{" ++ (texIncl v "Letter" i) ++ "}{" ++ showRuleName rName ++ (texShow (Cat (Letter c) r1)) ++ " \\sqsubseteq " ++ (texShow (Cat (Letter c) r2)) ++ "}"
texIncl v rName (LetterStar c r1 r2 r3 i) = "\\dfrac{" ++ (texIncl v "LetterStar" i) ++ "}{"  ++ showRuleName rName ++ (texShow (Cat (Letter c) r1)) ++ " \\sqsubseteq " ++ (texShow (Cat (Star r2) r3)) ++ "}"
texIncl v rName (LetterChoice2 c r1 r2 r3 r4 i) = "\\dfrac{" ++ (texIncl v "LetterChoice" i) ++ "}{"  ++ showRuleName rName ++ (texShow (Cat (Letter c) r1)) ++ "\\sqsubseteq " ++ (texShow (Cat (Choice r2 r3) r4)) ++ "}"
texIncl v rName (LetterChoice3 c r1 r2 r3 r4 i) = "\\dfrac{" ++ (texIncl v "LetterChoice" i) ++ "}{"  ++ showRuleName rName ++ (texShow (Cat (Letter c) r1)) ++ "\\sqsubseteq " ++ (texShow (Cat (Choice r2 r3) r4)) ++ "}"
texIncl v rName (LeftChoice r1 r2 r3 r4 i1 i2) = "\\dfrac{" ++ (texIncl v "LeftChoice" i1) ++ (texIncl v "" i2) ++ "}{" ++ showRuleName rName  ++ (texShow (Cat (Choice r1 r2) r3)) ++ " \\sqsubseteq " ++ (texShow r4) ++ "}"
texIncl v rName (LeftStar r1 r2 r3 r4 i1 i2) = "\\dfrac{" ++ (texIncl v "LeftStar" i1) ++ (texIncl v "" i2) ++ "}{" ++ showRuleName rName ++ (texShow (Cat (Star r1) r2)) ++ " \\sqsubseteq " ++ (texShow (Cat r3 r4)) ++ "}"
texIncl v rName (StarStarE r1 r2 i1) = "\\dfrac{" ++ (texIncl v "StarStarE" i1) ++ "}{"  ++ showRuleName rName ++ (texShow (Star r1)) ++ " \\sqsubseteq " ++ (texShow (Star r2)) ++ "}"
texIncl v rName (StarChoice1 r1 r2 r3 r4 r5 i1) = "\\dfrac{" ++ (texIncl v "StarChoice1" i1)  ++ "}{"  ++ showRuleName rName ++ (texShow (Cat (Star r1) r2)) ++ " \\sqsubseteq " ++ (texShow (Cat (Choice r3 r4) r5)) ++ "}"
texIncl v rName (StarChoice2 r1 r2 r3 r4 r5 i1 i2) = "\\dfrac{" ++ (texIncl v "StarChoice2" i1)  ++ "" ++ (texIncl v "" i2) ++ "}{"  ++ showRuleName rName ++ (texShow (Cat (Star r1) r2)) ++ " \\sqsubseteq " ++ (texShow (Cat (Choice r3 r4) r5)) ++ "}"
texIncl v rName (ElimCat r1 r2 r3 i1) = "\\dfrac{" ++ (texIncl v "ElimCat" i1)  ++ "}{" ++ showRuleName rName ++ (texShow r1) ++ " \\sqsubseteq " ++ (texShow (Cat r2 r3)) ++"}"
texIncl v rName (Loop r1 r2) = "\\dfrac{Store}{" ++ showRuleName rName ++ (texShow r1) ++ " \\sqsubseteq " ++ (texShow r2) ++ "}"
texIncl v rName (Not1Unambiguous s) = "\\text{Not 1-unambiguous}"
texIncl v rName (NotInclusion r1 r2 s) = "\\dfrac{}{" ++ showRuleName rName ++ (texShow r1) ++ "\\not \\sqsubseteq " ++ (texShow r2) ++ "}"

checkRule :: Inclusion -> Included
checkRule (NotInclusion _ _ _) = NotIncluded
checkRule (Not1Unambiguous _) = OneAmbiguous
checkRule (Loop _ _) = Included
checkRule (Axm _) = Included
checkRule (LetterRule _ _ _ i) = checkRule i
checkRule (LetterStar _ _ _ _ i) = checkRule i
checkRule (LetterChoice2 _ _ _ _ _ i) = checkRule i
checkRule (LetterChoice3 _ _ _ _ _ i) = checkRule i
checkRule (LeftChoice _ _ _ _ i1 i2) = combineIncluded (checkRule i1) (checkRule i2)
checkRule (LeftStar _ _ _ _ i1 i2) = combineIncluded (checkRule i1) (checkRule i2)
checkRule (StarStarE _ _ i) = checkRule i
checkRule (StarChoice1 _ _ _ _ _ i) = checkRule i
checkRule (StarChoice2 _ _ _ _ _ i1 i2) = combineIncluded (checkRule i1) (checkRule i2)
checkRule (ElimCat _ _ _ i) = checkRule i

combineIncluded :: Included -> Included -> Included
combineIncluded Included Included = Included
combineIncluded OneAmbiguous _ = OneAmbiguous
combineIncluded _ OneAmbiguous = OneAmbiguous
combineIncluded NotIncluded _ = NotIncluded
combineIncluded _ NotIncluded = NotIncluded



runInclusion :: Regexp -> Regexp -> Inclusion
runInclusion r1 r2 = fst (incFull [] r1 r2)

incFull :: [(Regexp, Regexp)] -> Regexp -> Regexp -> (Inclusion, [(Regexp, Regexp)])
incFull m r1 r2 = if nullable r1 && not (nullable r2) 
                  then (NotInclusion r1 r2 ("Not an inclusion, because " ++ (show r1) ++ " is nullable, while " ++ (show r2) ++ " is not."), m)
                  else incFull'' (List.find (\x -> x == (re1, re2)) m) ((re1,re2):m) re1 re2
    where re1 = (headerForm r1)
          re2 = (headerForm r2)

incFull'' :: Maybe (Regexp, Regexp) -> [(Regexp, Regexp)] -> Regexp -> Regexp -> (Inclusion, [(Regexp, Regexp)])
incFull'' (Just (r1,r2)) m _ _ = (Loop r1 r2, m)
incFull'' Nothing m r1 r2 = incFull' m r1 r2  

incFull' :: [(Regexp, Regexp)] -> Regexp -> Regexp -> (Inclusion, [(Regexp, Regexp)])

-- Axm
incFull' m Epsilon r = if nullable r then (Axm r, ((Epsilon, r):m)) else (NotInclusion Epsilon r ("Axm error: " ++ (show r) ++ " not nullable"), m)
--Letter
incFull' m (Cat (Letter l1) r1) (Cat (Letter l2) r2) = if (l1 == l2) 
                                                   then let (i1, m1) = incFull m r1 r2 in (LetterRule l1 r1 r2 i1, m1)
                                                   else (NotInclusion (Cat (Letter l1) r1) (Cat (Letter l2) r2) ("Cannot show that \"" ++ (show (Cat (Letter l1) r1)) ++ "\" is a subset of \"" ++ (show  (Cat (Letter l2) r2))), m)

-- LetterStar and ElimCat
incFull' m (Cat (Letter l) r1) (Cat (Star r2) r3) = 
    let {f2 = firstset r2; 
         f3 = firstset r3
        } in
    if elem l f2 
    then if elem l f3 
         then ( Not1Unambiguous 
                  ("Error while deciding \"" 
                   ++ (show (Cat (Letter l) r1)) 
                   ++ "\" <= \"" 
                   ++ (show (Cat (Star r2) r3)) 
                   ++ "\", since letter " 
                   ++ (show l) 
                   ++ " is in firstset of \"" 
                   ++ (show r2) ), m)
         else let (i1, m1) = (incFull m (Cat (Letter l) r1) (Cat r2 (Cat (Star r2) r3)))
              in (LetterStar l r1 r2 r3 i1, m1)
    else if elem l f3 
         then let (i1, m1) = (incFull m (Cat (Letter l) r1) r3) in (ElimCat (Cat (Letter l) r1) (Star r2) r3 i1, m1)
         else ( NotInclusion (Cat (Letter l) r1) (Cat (Star r2) r3) ("Cannot show that \"" 
                            ++ (show (Cat (Letter l) r1)) 
                            ++ "\" is a subset of \"" 
                            ++ (show  (Cat (Star r2) r3))), m)

                                                          
                                                           

-- LetterChoice and ElimCat
incFull' m (Cat (Letter l) r1) (Cat (Choice r2 r3) r4) = 
    let {f2 = firstset r2; 
         f3 = firstset r3;
         f4 = firstset r4
        } in
    if elem l f2
    then if elem l f3 
         then ( Not1Unambiguous ("Error while deciding \"" 
                               ++ (show (Cat (Letter l) r1)) 
                               ++ "\" <= \"" 
                               ++ (show (Cat (Choice r2 r3) r4)) 
                               ++ "\", since letter " 
                               ++ (show l) 
                               ++ " is in firstset of \"" 
                               ++ (show r3) ), m)
         else if (nullable (Choice r2 r3) && elem l f4) 
              then ( Not1Unambiguous ("Error while deciding \"" 
                                    ++ (show (Cat (Letter l) r1)) 
                                    ++ "\" <= \"" 
                                    ++ (show (Cat (Choice r2 r3) r4)) 
                                    ++ "\", since letter " 
                                    ++ (show l) 
                                    ++ " is in firstset of \"" 
                                    ++ (show r4) 
                                    ++ "\" and \"" 
                                    ++  (show r3) 
                                    ++ "\" is nullable."), m)
              else let (i1, m1) = (incFull m (Cat (Letter l) r1) (Cat r2 r4)) in (LetterChoice2 l r1 r2 r3 r4 i1, m1)
    else if elem l f3 
         then if nullable (Choice r2 r3)  && elem l f4 
              then ( Not1Unambiguous ("\"" 
                                    ++ (show (Cat (Choice r2 r3) r4)) 
                                    ++ "\" is 1-ambiguous"), m)
              else let (i1, m1) = (incFull m (Cat (Letter l) r1) (Cat r3 r4)) in (LetterChoice3 l r1 r2 r3 r4 i1, m1)
         else if nullable (Choice r2 r3) && elem l f4
              then let (i1, m1) = (incFull m (Cat (Letter l) r1) r4) in (ElimCat (Cat (Letter l) r1) (Choice r2 r3) r4 i1, m1)
              else ( NotInclusion (Cat (Letter l) r1) (Cat (Choice r2 r3) r4) ("Cannot show that \"" 
                                 ++ (show (Cat (Letter l) r1)) 
                                 ++ "\" is a subset of \"" 
                                 ++ (show  (Cat (Choice r2 r3) r4))), m)
                   
-- LeftChoice 
incFull' m (Cat (Choice r1 r2) r3) r4 = 
    let (i1, m1) = incFull m (Cat r1 r3) r4 in
    let (i2, m2) = incFull m1 (Cat r2 r3) r4 in (LeftChoice r1 r2 r3 r4 i1 i2, m2)
           

-- StarStarE
incFull' m (Cat (Star r1) Epsilon) (Cat (Star r2) Epsilon) = let (i1, m1) = (incFull m r1 (Cat (Star r2) Epsilon)) in (StarStarE r1 r2 i1, m1)
    

-- StarChoice1, StarChoice2 and ElimCat
incFull' m (Cat (Star r1) r2) (Cat (Choice r3 r4) r5) = let {f1 = firstset r1; 
                                                             f2 = firstset r2; 
                                                             f3 = firstset r3; 
                                                             f4 = firstset r4; 
                                                             f35 = firstset (Cat r3 r5);
                                                             f45 = firstset (Cat r4 r5);
                                                             f5 = firstset r5
                                                            } in
                                                        if (hasIntersection f1 f3 || hasIntersection f2 f3)
                                                               && (subset f1 f35) 
                                                               && (subset f2 f35) 
                                                               && ((not(nullable r2)) || (nullable r3))
                                                        then if hasIntersection f1 f4 || hasIntersection f2 f4 
                                                             then ( Not1Unambiguous ("Not 1-unambiguous: " ++ (show (Cat (Choice r3 r4) r5))) , m )
                                                             else let (i1, m1) = (incFull m (Cat (Star r1) r2) (Cat r3 r5)) in (StarChoice1 r1 r2 r3 r4 r5 i1, m1)
                                                        else if (hasIntersection f1 f4 || hasIntersection f2 f4)
                                                                 && (subset f1 f45) 
                                                                 && (subset f2 f45) 
                                                                 && ((not(nullable r2)) || (nullable r4))
                                                             then if hasIntersection f1 f3 || hasIntersection f2 f3 
                                                                  then ( Not1Unambiguous ("Not 1-unambiguous: " ++ (show (Cat (Choice r3 r4) r5))) , m)
                                                                  else let (i1, m1) = (incFull m (Cat (Star r1) r2) (Cat r4 r5)) in (StarChoice1 r1 r2 r3 r4 r5 i1, m1)
                                                             else if (subset f1 f5) && (subset f2 f5) && (nullable (Choice r3 r4))
                                                                  then if hasIntersection f1 f3 
                                                                           || hasIntersection f1 f4 
                                                                           || hasIntersection f2 f3 
                                                                           || hasIntersection f2 f4
                                                                       then ( Not1Unambiguous ((show (Cat (Choice r3 r4) r5)) ++ " is not 1-unambiguous"), m)
                                                                       else let (i1, m1) = (incFull m (Cat (Star r1) r2) r5) in (ElimCat (Cat (Star r1) r2) (Choice r3 r4) r5 i1, m1)
                                                                  else let (i2, m1) = (incFull m r2 (Cat (Choice r3 r4) r5)) 
                                                                       in let (i1, m2) = (incFull m1 (Cat r1 (Cat (Star r1) r2)) (Cat (Choice r3 r4) r5))
                                                                          in (StarChoice2 r1 r2 r3 r4 r5 i1 i2, m2)
                                                                           
                                                                           


                                    
-- LeftStar and ElimCat
-- Note that r3 cannot be a choice, since starchoice is above, 
-- and r3 cannot be starred if r4 is epsilon, since starstare is above
incFull' m (Cat (Star r1) r2) (Cat r3 r4) = 
    let {f1 = firstset r1; 
         f3 = firstset r3;
         f4 = firstset r4
        } in
    if hasIntersection f1 f3 
    then if nullable r3 && subset f1 f4 
         then ( Not1Unambiguous ("Firstset of \"" 
                               ++ (show r1) 
                               ++ "\" is included in firstset of \"" 
                               ++ (show r4) 
                               ++ ". This came up in the decision \"" 
                               ++ (show (Cat (Star r1) r2)) 
                               ++ "\" <= \"" 
                               ++ (show  (Cat r3 r4)) 
                               ++ "\"."), m)
         else if r4 == Epsilon && Inclusion.isLetter r3
              then ( NotInclusion (Cat (Star r1) r2) r3 ("LeftStar prevents showing " ++ (show (Cat (Star r1) r2)) ++ " <= " ++ (show r3) ++ ". A single letter cannot be the superset of a starred expression."), m)
              else let (i1, m1) = (incFull m (Cat r1 (Cat (Star r1) r2)) (Cat r3 r4)) in
                   let (i2, m2) = (incFull m1 r2 (Cat r3 r4))
                   in (LeftStar r1 r2 r3 r4 i1 i2, m2)
    else if (subset (firstset r1) (firstset r4))
         then let (i1, m1) = (incFull m (Cat (Star r1) r2) r4) in (ElimCat (Cat (Star r1) r2) r3 r4 i1, m1)
         else (NotInclusion (Cat (Star r1) r2) (Cat r3 r4) ("Cannot show that \"" ++ (show (Cat (Star r1) r2)) ++ "\" is a subset of \"" ++ (show  (Cat r3 r4))  ++ "\""), m)

                                                                                  
                                                             

incFull' _ r1 r2 = (NotInclusion r1 r2 ("Error in call to incFull': " ++ show r1 ++ "<=" ++ show r2 ++ "\n"), [])

                                                                                 
inclusionT :: String -> String -> Bool -> IO ()
inclusionT r1 r2 verbose = inclusionVerbose r1 r2 verbose (texIncl True "")
            

inclusionVerbose :: String -> String -> Bool -> (Inclusion -> String) -> IO ()
inclusionVerbose r1 r2 verbose printer = showAll (runInclusion (parseRE r1) (parseRE r2))
    where showAll res = do
            putStrLn ("\n" ++ (show (checkRule res)) ++ "\n")
            showVerbose res verbose
              where
                showVerbose _ False = do putStrLn "\n"
                showVerbose result True = do putStrLn (texIncl False "" result)
                                             putStrLn (prettyIncl result)

inclusionT2 :: String -> String -> Included
inclusionT2 r1 r2 = checkRule (runInclusion (parseRE r1) (parseRE r2))
                  
inclusionTA :: [String] -> IO ()
inclusionTA [r1,r2] = inclusionT r1 r2 False
inclusionTA ["-v",r1,r2] = inclusionT r1 r2 True
inclusionTA _ = error "Usage: <prog> [-v] <re1> <re2>"
   
